//**********************************************************************
// Name:	SiSmDriver.c
// Description:	
//**********************************************************************

#include "SiSmDriver.h"

static EFI_GUID gDriverBindingProtocolGuid		= EFI_DRIVER_BINDING_PROTOCOL_GUID;
static EFI_GUID gEfiComponentNameProtocolGuid		= EFI_COMPONENT_NAME_PROTOCOL_GUID;
static EFI_GUID gDevicePathProtocolGuid			= EFI_DEVICE_PATH_PROTOCOL_GUID;

EFI_GUID 	gEfiSiSmIoProtocolGuid 			= EFI_SISM_IO_PROTOCOL_GUID;

#define END_DEVICE_PATH			0x7F
#define END_DEVICE_PATH1		0xFF
#define END_INSTANCE_SUBTYPE		0x01
#define END_ENTIRE_SUBTYPE		0xFF

#define NODE_LENGTH(pPath) (*(UINT16*)&(pPath)->Length[0])
#define SET_NODE_LENGTH(pPath,Length) (NODE_LENGTH(pPath)=(Length))
#define NEXT_NODE(pPath) ((EFI_DEVICE_PATH_PROTOCOL*)((UINT8*)(pPath)+NODE_LENGTH(pPath)))
#define isEndNode(pDp) ((pDp)->Type==END_DEVICE_PATH||(pDp)->Type==END_DEVICE_PATH1)

// EFI Event Types
#define EFI_EVENT_TIMER                         0x80000000
#define EFI_EVENT_NOTIFY_SIGNAL                 0x00000200


// Function Prototypes
EFI_STATUS SiSmIoSetValueOn (
  VOID
);

EFI_STATUS SiSmIoSetValueOff (
  VOID
);

EFI_STATUS SiSmIoSetValueDec (
  VOID
);

EFI_SISM_IO_PROTOCOL SiSmIoProtocol = { 
    SiSmIoSetValueOn,
    SiSmIoSetValueOff,
    SiSmIoSetValueDec,
	0,
	0,
	0
};

static 	UINT16		*gDriverName=L"Hello EFI-Driver world";

static 	UINT16		iNum = 0;
EFI_SYSTEM_TABLE 	*gSysTable = NULL;
EFI_EVENT   		gEvSiSmIoTimer =0;
UINTN			gSiSmDrvCtl = 1;//2;
EFI_SISM_DEVICE		*gSiSmDevice;

//==================================================================================
//Function Prototypes for Driver Binding Protocol Interface
//==================================================================================
EFI_STATUS SiSmDriverSupported(IN EFI_DRIVER_BINDING_PROTOCOL	*This,
				IN EFI_HANDLE                    Controller,
				IN EFI_DEVICE_PATH_PROTOCOL      *RemainingDevicePath);

EFI_STATUS SiSmDriverStart(IN EFI_DRIVER_BINDING_PROTOCOL		*This,
					   IN EFI_HANDLE		Controller,
					   IN EFI_DEVICE_PATH_PROTOCOL	*RemainingDevicePath );

EFI_STATUS SiSmDriverStop(IN EFI_DRIVER_BINDING_PROTOCOL	*This,
					   IN EFI_HANDLE	Controller,
					   IN UINTN		NumberOfChildren,
					   IN EFI_HANDLE	*ChildHandleBuffer);

EFI_STATUS SiSmDriverComponentNameGetControllerName(IN EFI_COMPONENT_NAME_PROTOCOL *This,
						IN  EFI_HANDLE                   ControllerHandle,
						IN  EFI_HANDLE                   ChildHandle        OPTIONAL,
						IN  CHAR8                        *Language,
						OUT CHAR16                       **ControllerName );

EFI_STATUS SiSmDriverComponentNameGetDriverName(IN  EFI_COMPONENT_NAME_PROTOCOL  *This,
						IN  CHAR8                        *Language,
						OUT CHAR16                       **DriverName);

//==================================================================================
//Driver binding protocol instance for SiSmDriver Driver
EFI_DRIVER_BINDING_PROTOCOL gSiSmDriverDriverBinding = {
	SiSmDriverSupported,	        //Supported
	SiSmDriverStart,		//,
	SiSmDriverStop,			//,
	0x1,				//Version,
	NULL,				//Image Handle,
	NULL				//DriverBindingHandle,
};

//==================================================================================
//Component Name Protocol Instance
EFI_COMPONENT_NAME_PROTOCOL gSiSmDriverComponentName = {
  SiSmDriverComponentNameGetDriverName,
  SiSmDriverComponentNameGetControllerName,
  "eng"
};

//EFI_COMPONENT_NAME_PROTOCOL gSiSmDriverDriverName;

//==============================================================================

EFI_STATUS SiSmDevicePath(EFI_DEVICE_PATH_PROTOCOL *DevicePath)
{
	EFI_STATUS	Status = EFI_UNSUPPORTED;
//	CHAR16		Buffer16[512];
	UINTN		Count = 0;
 
	for(;;)
	{
		switch(DevicePath->Type)
		{
		case HARDWARE_DEVICE_PATH:
			switch(DevicePath->SubType)
			{
			case HW_PCI_DP:
				Print(L"PCI(%X|%X)\\",((PCI_DEVICE_PATH*)DevicePath)->Device,((PCI_DEVICE_PATH*)DevicePath)->Function);
				break;
			case HW_PCCARD_DP:
				//Print(L"Pccard(Socket %i)\\",((PCCARD_DEVICE_PATH*)DevicePath)->Function);
				break;
			case HW_VENDOR_DP:
				Print(L"VenHw(%g)\\",&((VENDOR_DEVICE_PATH*)DevicePath)->Guid);
  				if( ((VENDOR_DEVICE_PATH*)DevicePath)->Guid.Data1 == 0x4E11E955)
					Status = EFI_SUCCESS;
				break;
			case HW_MEMMAP_DP:
				Print(L"MemMap(Type %i, SubType %i)",DevicePath->Type,DevicePath->SubType);	
				Count = 100;//Exit
				break;
			default:Print(L"DevicePath(Type %i, SubType %i)????",DevicePath->Type,DevicePath->SubType);	
				Count = 100;//Exit
			}		
			break;
		case ACPI_DEVICE_PATH:
				switch(DevicePath->SubType)
				{
				case ACPI_DP:
					Print(L"Acpi(%x, %x)\\", ((ACPI_HID_DEVICE_PATH*)DevicePath)->HID, ((ACPI_HID_DEVICE_PATH*)DevicePath)->UID);
//  					if(((ACPI_HID_DEVICE_PATH*)DevicePath)->HID == EISA_PNP_ID(0x303)  && ((ACPI_HID_DEVICE_PATH*)DevicePath)->UID == 0)
						Status = EFI_SUCCESS;
					break;
				default:Print(L"DevicePath(Type %i, SubType %i)????",DevicePath->Type,DevicePath->SubType);
				Count = 100;//Exit
				}
			break;
		case MESSAGING_DEVICE_PATH:
			switch(DevicePath->SubType)
			{
			case MSG_ATAPI_DP:
				if (((ATAPI_DEVICE_PATH*)DevicePath)->PrimarySecondary == 0){
				Print(L"ATA(%s, %i)\\", "Master",((ATAPI_DEVICE_PATH*)DevicePath)->Lun);
					break;
					}
				else if (((ATAPI_DEVICE_PATH*)DevicePath)->PrimarySecondary == 1){
					Print(L"ATA(%s, %i)\\", L"Slave", ((ATAPI_DEVICE_PATH*)DevicePath)->Lun);
					break;
					}
				else break;
			case MSG_SCSI_DP:
				Print(L"SCSI(%i, %i)\\", ((SCSI_DEVICE_PATH*)DevicePath)->Pun, ((SCSI_DEVICE_PATH*)DevicePath)->Lun);
				break;
			case MSG_VENDOR_DP:
				Print(L"VenMsg(%g)\\", ((VENDOR_DEVICE_PATH*)DevicePath)->Guid);
				break;
			case MSG_USB_DP:
				Print(L"USB(%x,%x)\\",	((USB_DEVICE_PATH*)DevicePath)->ParentPortNumber,((USB_DEVICE_PATH*)DevicePath)->InterfaceNumber);
				break;
			default:Print(L"DevicePath(Type %i, SubType %i)??????",DevicePath->Type,DevicePath->SubType);	
				Count = 100;//Exit
			}
			break;
		case MEDIA_DEVICE_PATH:
			switch(DevicePath->SubType)
			{
			case MEDIA_HARDDRIVE_DP:
			//	TODO: Find out hard drive signature format to display.
				Print(L"HD(Part%i, Sig %s)\\", ((HARDDRIVE_DEVICE_PATH*)DevicePath)->PartitionNumber, "?");// ((HARDDRIVE_DEVICE_PATH*)Path)->Signature);
				break;
			case MEDIA_CDROM_DP:
				Print(L"CDROM(Entry%i)\\", ((CDROM_DEVICE_PATH*)DevicePath)->BootEntry);
				break;
			case MEDIA_VENDOR_DP:
				Print(L"VenMedia(%g)\\", ((VENDOR_DEVICE_PATH*)DevicePath)->Guid);
				break;
			case MEDIA_FILEPATH_DP:
				//Print(L"%g\\", &((MEDIA_FW_VOL_FILEPATH_DEVICE_PATH*)DevicePath)->NameGuid);
				break;
			default:Print(L"DevicePath(Type %i, SubType %i)?????",DevicePath->Type,DevicePath->SubType);
				Count = 100;//Exit
			}
			break;
		case END_DEVICE_PATH:
		        Print(L"/END_PATH/ " );
			break;
		case END_DEVICE_PATH1:
		        Print(L" END_PATH1 " );
			Count = 100;//Exit
			break;
		default:
			Print(L"DevicePath(Type %i, SubType %i)?????",DevicePath->Type,DevicePath->SubType);
			Count = 100;//Exit
		}
		if(Status == EFI_SUCCESS)gSysTable->ConOut->OutputString(gSysTable->ConOut, L" Ok!" );
		if (Count++ >= 50) break;
		if (isEndNode(DevicePath) && DevicePath->SubType == END_ENTIRE_SUBTYPE) break;
		
		DevicePath = NEXT_NODE(DevicePath);
	}
	gSysTable->ConOut->OutputString(gSysTable->ConOut,L":" ); 

	return Status;
}

//---------------------------------------------------------------------------
//
// Name:        SiSmIoOnTimer
//
// Description: Timer call-back routine that 
//              is used to monitore changes Port 0x80
//
//---------------------------------------------------------------------------

VOID
SiSmIoOnTimer(
    EFI_EVENT   Event,
    VOID        *Context
)
{
	static UINT8	ValPort80=0;

	switch(	gSiSmDrvCtl ){
		case 1: 
                        ValPort80++ ;
			break;
		case 2: ValPort80-- ;
			break;
		default:
			break;
	}
	if(gSiSmDrvCtl){
	        UINTN 	X,Y, MaxX, MaxY;

	        gSysTable->ConOut->QueryMode(gSysTable->ConOut,gSysTable->ConOut->Mode->Mode, &MaxX, &MaxY);
	        //Get cursor position
	        X = gSysTable->ConOut->Mode->CursorColumn;
	        Y = gSysTable->ConOut->Mode->CursorRow;
	        //Set cursor and print string
	        gSysTable->ConOut->SetCursorPosition(gSysTable->ConOut, MaxX-10, MaxY-1);
	       	Print(L" %x ", ValPort80 );
	        //Restore cursor position
	  	gSysTable->ConOut->SetCursorPosition (gSysTable->ConOut,  X, Y);
	}

        return;
}
//---------------------------------------------------------------------------
//
// Name:	SiSmDriverEntryPoint
//
// Description:	
//
// Output:     
//
//---------------------------------------------------------------------------
EFI_STATUS
EFIAPI
SiSmDriverEntryPoint (
  IN EFI_HANDLE           ImageHandle,
  IN EFI_SYSTEM_TABLE     *SystemTable
  )
{
	EFI_STATUS				Status = EFI_SUCCESS;
	
  	// Initialize the EFI library
//  	InitAmiLib(ImageHandle, SystemTable);
	InitializeLib(ImageHandle, SystemTable);//EDKII

	// Initialize global variable for use in this driver
    	gSysTable = SystemTable;
	
	//Fill ImgHandle Field in Driver Binding Protocol Structure
	gSiSmDriverDriverBinding.ImageHandle=ImageHandle;
	gSiSmDriverDriverBinding.DriverBindingHandle=NULL;
	gSysTable->ConOut->OutputString(gSysTable->ConOut, L"  _____ SiSmDriverEntryPoint ____ \n\r" ); 
 
	//install driver binding protocol here ....
	Status = gSysTable->BootServices->InstallMultipleProtocolInterfaces(
					&gSiSmDriverDriverBinding.DriverBindingHandle,
					&gDriverBindingProtocolGuid,&gSiSmDriverDriverBinding,
					&gEfiComponentNameProtocolGuid, &gSiSmDriverComponentName,
					NULL);

	//Here we can set up notification events if needed

	//TRACE((-1,"\nSiSmDriverEntryPoint: Ende\n"));
  	return Status;
}


//==================================================================================
//Functions for Driver Binding Protocol Interface
//==================================================================================
EFI_STATUS 
SiSmDriverSupported(		IN EFI_DRIVER_BINDING_PROTOCOL	*This,
				   	IN EFI_HANDLE                    Controller,
					IN EFI_DEVICE_PATH_PROTOCOL      *RemainingDevicePath)
{
	EFI_STATUS				Status 	= EFI_SUCCESS;
	EFI_STATUS				StatusDP = EFI_UNSUPPORTED;
 	EFI_DEVICE_PATH_PROTOCOL		*pDP;
	BOOLEAN					AlreadyStarted = FALSE;
	EFI_SISM_IO_PROTOCOL			*SiSmIo;
	static UINTN				iNum=0;

        Print(L"  __Support__  %x ", iNum++ );
	//gSysTable->ConOut->OutputString(gSysTable->ConOut, L"Support\n\r" ); 

	//TRACE((-1,"\nSiSmDriverSupported:\n"));

	//Checking if Protocol "SiSmIo" is started ?
	Status = gSysTable->BootServices->OpenProtocol (Controller, 
							&gEfiSiSmIoProtocolGuid,	//
    	  						&SiSmIo,			//
							This->DriverBindingHandle, 	//
							Controller, 
							EFI_OPEN_PROTOCOL_BY_DRIVER );

        if ((Status == EFI_ALREADY_STARTED)||(Status == EFI_SUCCESS)) {
		//Ok Close checked Protocol 
    		gSysTable->BootServices->CloseProtocol (	Controller, &gEfiSiSmIoProtocolGuid,
	   							This->DriverBindingHandle, Controller);
		gSysTable->ConOut->OutputString(gSysTable->ConOut, L"Protocol already Installed:EFI_UNSUPPORTED" );
     	   	return EFI_UNSUPPORTED;
    	}

	// Get device path from Controller handle.
	Status = gSysTable->BootServices->OpenProtocol (
	                Controller,
	                &gDevicePathProtocolGuid,
	                (VOID**)&pDP,
	                This->DriverBindingHandle,
	                Controller,   
	                EFI_OPEN_PROTOCOL_BY_DRIVER
	                );
    	if (EFI_ERROR(Status)) {
		gSysTable->ConOut->OutputString(gSysTable->ConOut, L"  __Unsupported__:OpenProtocol \n\r" ); 
        	return EFI_UNSUPPORTED;
    	}

	gSysTable->BootServices->CloseProtocol(Controller,&gDevicePathProtocolGuid,This->DriverBindingHandle, Controller);

	StatusDP = SiSmDevicePath(pDP);// for EDKII

//    	ps2DevPath=DPGetLastNode(pDP); // for AMI 

    	// ps2DevPath is now the last node
//    	acpiDP = (ACPI_HID_DEVICE_PATH*)ps2DevPath;
//    	StatusDP = (acpiDP->Header.Type == ACPI_DEVICE_PATH && 
//        	acpiDP->Header.SubType == ACPI_DP)? EFI_SUCCESS : EFI_UNSUPPORTED;

	if(StatusDP != EFI_SUCCESS){
		gSysTable->ConOut->OutputString(gSysTable->ConOut, L"  __Unsupported__:DP \n\r" ); 
		return EFI_UNSUPPORTED;
	}

	//TRACE((-1,"\nSiSmDriverSupported:  Controller=0%x\n", Controller));

	Status = gSysTable->BootServices->AllocatePool(	EfiBootServicesData,		//enum
							sizeof(EFI_SISM_DEVICE),
							&gSiSmDevice);
	
	gSysTable->BootServices->SetMem(gSiSmDevice, sizeof(EFI_SISM_DEVICE),0);

	gSiSmDevice->Signature		= EFI_SISM_DEVICE_SIGNATURE;
	gSiSmDevice->SiSmIoProtocol	= SiSmIoProtocol;	
	gSiSmDevice->PrivateData1	= 1;
	gSiSmDevice->PrivateData2	= 2;
	gSiSmDevice->SiSmIoProtocol.Rev 	= EFI_SISM_IO_PROTOCOL_REVISION;

    	// Instal SISM_IO_PROTOCOL

//    	VERIFY_EFI_ERROR(
        	gSysTable->BootServices->InstallProtocolInterface( &Controller,
            	&gEfiSiSmIoProtocolGuid, EFI_NATIVE_INTERFACE, &gSiSmDevice->SiSmIoProtocol );//EDKII );

	Status = gSysTable->BootServices->OpenProtocol (Controller, 
							&gEfiSiSmIoProtocolGuid,	
    	  						&SiSmIo,			
							This->DriverBindingHandle, 	
							Controller, 
							EFI_OPEN_PROTOCOL_BY_DRIVER );
	gSysTable->ConOut->OutputString(gSysTable->ConOut, L"  __Supported__:Ok! \n\r" ); 

 	Status=EFI_SUCCESS;
	return Status;
}

EFI_STATUS 
SiSmDriverStart(IN EFI_DRIVER_BINDING_PROTOCOL	*This,
			   IN EFI_HANDLE			Controller,
			   IN EFI_DEVICE_PATH_PROTOCOL		*RemainingDevicePath )
{
	
	EFI_STATUS			Status = EFI_SUCCESS;

	gSysTable->ConOut->OutputString(gSysTable->ConOut, L"  __.. SiSmDriver Start  \n\r" ); 

    	// Install Polling timer
    	{
        	EFI_TPL SaveTpl = gSysTable->BootServices->RaiseTPL (EFI_TPL_CALLBACK);

        	ASSERT( SaveTpl <= EFI_TPL_CALLBACK);

        	// Critical section

		//TRACE((-1,"\n...Install Polling timer\n"));

//        	VERIFY_EFI_ERROR(
                    	gSysTable->BootServices->CreateEvent ( EFI_EVENT_TIMER | EFI_EVENT_NOTIFY_SIGNAL, 
						EFI_TPL_NOTIFY, SiSmIoOnTimer,0,&gEvSiSmIoTimer);//EDKII );  
//                	VERIFY_EFI_ERROR(
                    		gSysTable->BootServices->SetTimer ( gEvSiSmIoTimer, TimerPeriodic, ONESECOND/2);//EDKII);
										//TimerPeriodic: enum
        	gSysTable->BootServices->RestoreTPL(SaveTpl);
    	}
	return Status;
}

EFI_STATUS SiSmDriverStop(IN EFI_DRIVER_BINDING_PROTOCOL	*This,
			   IN EFI_HANDLE			Controller,
			   IN UINTN				NumberOfChildren,
			   IN EFI_HANDLE			*ChildHandleBuffer)
{
	EFI_STATUS	Status=EFI_SUCCESS;

	EFI_TPL SaveTpl = gSysTable->BootServices->RaiseTPL (EFI_TPL_CALLBACK);
	ASSERT(SaveTpl <= EFI_TPL_CALLBACK );

//        VERIFY_EFI_ERROR(
            gSysTable->BootServices->SetTimer ( gEvSiSmIoTimer, TimerCancel, ONESECOND/10);//EDKII );
//        VERIFY_EFI_ERROR(
            gSysTable->BootServices->CloseEvent ( gEvSiSmIoTimer );//EDKII );

	if( gSiSmDevice != NULL) 
		gSysTable->BootServices->FreePool( gSiSmDevice );

	//TRACE((-1,"\nSiSmDriverStop:  Controller=0x\n", Controller));

	gSysTable->BootServices->RestoreTPL(SaveTpl);

    	return Status;
}

//==================================================================================
//Functions for Driver Binding Protocol Interface
//==================================================================================
EFI_STATUS SiSmDriverComponentNameGetControllerName(IN EFI_COMPONENT_NAME_PROTOCOL *This,
						IN  EFI_HANDLE                   ControllerHandle,
						IN  EFI_HANDLE                   ChildHandle        OPTIONAL,
						IN  CHAR8                        *Language,
						OUT CHAR16                       **ControllerName )
{
	//TRACE((-1,"\nSiSmDriverComponentNameGetControllerName: \n"));
	gSysTable->ConOut->OutputString(gSysTable->ConOut, L"\n\r  __...SiSmDriver Component Name/GetControllerNam  \n\r" ); 

	return EFI_UNSUPPORTED;
}

EFI_STATUS SiSmDriverComponentNameGetDriverName(IN  EFI_COMPONENT_NAME_PROTOCOL  *This,
						IN  CHAR8                        *Language,
						OUT CHAR16                       **DriverName)
{
	UINT8	*l="eng";

	//TRACE((-1,"\n SiSmDriverComponentNameGetDriverName: \n"));
	gSysTable->ConOut->OutputString(gSysTable->ConOut, L" __ SiSmDriver Get Driver Name  " ); 

	//Supports only "eng" 
	if(!Language || !DriverName) return EFI_INVALID_PARAMETER;

	if( Language[0] != l[0] )
                return EFI_UNSUPPORTED;
	
	else *DriverName=gDriverName;
	
	return EFI_SUCCESS;
}

EFI_STATUS 
EFIAPI
EFIAPI SiSmIoSetValueOn (  VOID )
{
  	EFI_STATUS    		Status=EFI_SUCCESS;

	gSiSmDrvCtl = 1; 
	return Status;
}

EFI_STATUS 
EFIAPI
EFIAPI SiSmIoSetValueOff (  VOID )
{
  	EFI_STATUS    		Status=EFI_SUCCESS;

	gSiSmDrvCtl = 0; 
	return Status;
}

EFI_STATUS 
EFIAPI
EFIAPI SiSmIoSetValueDec (  VOID )
{
  	EFI_STATUS    		Status=EFI_SUCCESS;

	gSiSmDrvCtl = 2; 
	return Status;
}
